"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.plugin = exports.FunctionBuilder = void 0;
const path_1 = __importDefault(require("path"));
const archiver_1 = __importDefault(require("archiver"));
const fs_1 = __importDefault(require("fs"));
const framework_core_1 = require("@cloudbase/framework-core");
const useSAMDeploy = false;
class FunctionPlugin extends framework_core_1.Plugin {
    constructor(name, api, inputs) {
        super(name, api, inputs);
        this.name = name;
        this.api = api;
        this.inputs = inputs;
        const config = this.api.projectConfig;
        const DEFAULT_INPUTS = {
            functionRootPath: (config === null || config === void 0 ? void 0 : config.functionRoot) || "cloudfunctions",
            functions: config === null || config === void 0 ? void 0 : config.functions,
            servicePaths: {},
        };
        this.resolvedInputs = resolveInputs(this.inputs, DEFAULT_INPUTS);
        this.resolvedInputs.functions = this.resolvedInputs.functions.map((func) => {
            return Object.assign({}, {
                runtime: "Nodejs10.15",
                installDependency: true,
                handler: "index.main",
            }, func);
        });
        this.functions = this.resolvedInputs.functions;
        this.functionRootPath = path_1.default.isAbsolute(this.resolvedInputs.functionRootPath)
            ? this.resolvedInputs.functionRootPath
            : path_1.default.join(this.api.projectPath, this.resolvedInputs.functionRootPath);
        this.builder = new FunctionBuilder({
            projectPath: this.api.projectPath,
        });
        this.outputs = {};
    }
    /**
     * 初始化
     */
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            this.api.logger.debug("FunctionPlugin: init", this.resolvedInputs);
        });
    }
    compile() {
        return __awaiter(this, void 0, void 0, function* () {
            this.api.logger.debug("FunctionPlugin: compile", this.resolvedInputs);
            if (useSAMDeploy) {
                const builderOptions = this.functions.map((func) => {
                    const localFunctionPath = path_1.default.join(this.functionRootPath, func.name);
                    const zipName = `${func.name + Date.now()}.zip`;
                    return {
                        name: func.name,
                        localPath: localFunctionPath,
                        zipfileName: zipName,
                    };
                });
                const buildResult = yield this.builder.build(builderOptions);
                yield Promise.all(buildResult.functions.map((func) => __awaiter(this, void 0, void 0, function* () {
                    const cloudPath = `framework-upload/${func.name}.zip`;
                    const url = yield this.uploadToCos(func.source, cloudPath);
                    this.outputs[func.name] = {
                        codeUrl: url,
                    };
                })));
            }
            return {
                EntryPoint: Object.values(this.resolvedInputs.servicePaths).map((servicePath) => {
                    return {
                        Label: "服务地址",
                        EntryType: "HttpService",
                        HttpEntryPath: servicePath,
                    };
                }),
                Resources: this.functions.reduce((resources, func) => {
                    resources[this.toConstantCase(func.name)] = this.functionConfigToSAM(func);
                    return resources;
                }, {}),
            };
        });
    }
    /**
     * 执行本地命令
     */
    run() {
        return __awaiter(this, void 0, void 0, function* () { });
    }
    /**
     * 删除资源
     */
    remove() {
        return __awaiter(this, void 0, void 0, function* () { });
    }
    /**
     * 生成代码
     */
    genCode() {
        return __awaiter(this, void 0, void 0, function* () { });
    }
    /**
     * 构建
     */
    build() {
        return __awaiter(this, void 0, void 0, function* () {
            this.api.logger.debug("FunctionPlugin: build", this.resolvedInputs);
        });
    }
    /**
     * 部署
     */
    deploy() {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            this.api.logger.debug("FunctionPlugin: deploy", this.resolvedInputs, this.buildOutput);
            const Function = (_a = this.api.resourceProviders) === null || _a === void 0 ? void 0 : _a.function;
            // 批量部署云函数
            yield Promise.all(this.functions.map((func) => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield Function.createFunction({
                        func,
                        envId: this.api.envId,
                        force: true,
                        functionRootPath: this.functionRootPath,
                    });
                    this.api.logger.info(`${this.api.emoji("🚀")} [${func.name}] 云函数部署成功`);
                }
                catch (e) {
                    this.api.logger.error(`${this.api.emoji("🙅‍♂")} [${func.name}] 函数部署失败`);
                    throw new Error(e.message);
                }
            })));
            // 批量处理云接入
            yield Promise.all(Object.entries(this.resolvedInputs.servicePaths).map(([functionName, servicePath]) => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield this.api.cloudbaseManager.commonService().call({
                        Action: "CreateCloudBaseGWAPI",
                        Param: {
                            ServiceId: this.api.envId,
                            Path: servicePath,
                            Type: 1,
                            Name: functionName,
                        },
                    });
                }
                catch (e) {
                    if (!e.message.includes("api created")) {
                        throw e;
                    }
                }
                let url = `https://${this.api.envId}.service.tcloudbase.com${servicePath}`;
                if (url[url.length - 1] !== "/") {
                    url = url + "/";
                }
                url = this.api.genClickableLink(url);
                this.api.logger.info(`${this.api.emoji("🚀")} 云接入服务发布成功，访问地址: ${url}`);
            })));
            this.api.logger.info(`${this.api.emoji("🚀")} 云函数部署成功`);
        });
    }
    functionConfigToSAM(functionConfig) {
        return {
            Type: "CloudBase::Function",
            Properties: {
                Handler: functionConfig.handler || "index.main",
                Description: "CloudBase Framework 部署的云函数",
                Runtime: functionConfig.runtime,
                FunctionName: functionConfig.name,
                MemorySize: functionConfig.memory || 128,
                Timeout: functionConfig.timeout || 5,
                Environment: functionConfig.envVariables,
                VpcConfig: functionConfig.vpc,
                HttpPath: this.resolvedInputs.servicePaths[functionConfig.name],
                InstallDependency: "installDependency" in functionConfig
                    ? functionConfig.installDependency
                    : true,
                CodeUri: this.outputs[functionConfig.name] &&
                    this.outputs[functionConfig.name].codeUrl,
                Role: "TCB_QcsRole",
            },
        };
    }
    toConstantCase(name) {
        let result = "";
        let lastIsDivide = true;
        for (let i = 0; i < name.length; i++) {
            let letter = name[i];
            if (letter === "-" || letter === "_") {
                lastIsDivide = true;
            }
            else if (lastIsDivide) {
                result += letter.toUpperCase();
                lastIsDivide = false;
            }
            else {
                result += letter.toLowerCase();
                lastIsDivide = false;
            }
        }
        return result;
    }
    uploadToCos(localPath, cloudPath) {
        return __awaiter(this, void 0, void 0, function* () {
            // @todo use cloudId
            const uploadResult = yield this.api.cloudbaseManager.storage.uploadFile({
                localPath,
                cloudPath,
            });
            const result = yield this.api.cloudbaseManager.storage.getTemporaryUrl([
                {
                    cloudPath,
                    maxAge: 86400,
                },
            ]);
            return result[0].url;
        });
    }
}
function resolveInputs(inputs, defaultInputs) {
    return Object.assign({}, defaultInputs, inputs);
}
class FunctionBuilder extends framework_core_1.Builder {
    constructor(options) {
        super(Object.assign({ type: "function" }, options));
    }
    build(options) {
        return __awaiter(this, void 0, void 0, function* () {
            return {
                functions: options.map((option) => {
                    const localZipPath = path_1.default.join(this.distDir, option.zipfileName);
                    if (!fs_1.default.existsSync(this.distDir)) {
                        fs_1.default.mkdirSync(this.distDir, { recursive: true });
                    }
                    this.zipDir(option.localPath, localZipPath);
                    return {
                        name: option.name,
                        options: {},
                        source: localZipPath,
                        entry: option.zipfileName,
                    };
                }),
            };
        });
    }
    zipDir(src, dest) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                // create a file to stream archive data to.
                var output = fs_1.default.createWriteStream(dest);
                var archive = archiver_1.default("zip", {
                    zlib: { level: 9 },
                });
                output.on("close", resolve);
                archive.on("error", reject);
                archive.directory(src, false);
                archive.pipe(output);
                archive.finalize();
            });
        });
    }
}
exports.FunctionBuilder = FunctionBuilder;
exports.plugin = FunctionPlugin;
